#include "eigen.h"

/* Eigen is the most important routine in pori concerning eigenvalues and
   eigenvectors. It takes as input a symmetric matrix, which MUST be allocated
   as a whole matrix, and should contain values only in the lower triangle
   (where the second index is lower or equal to the first). It also takes as
   an input the size of the matrix. Then it takes three more inputs - one
   a vector of length n (for the eigenvalues) and the second is a rectangular
   matrix of size nXs, and the last specifies n - the number of eigenvectors
   and eigen values to find. The n biggest eigenvalues and their respective
   eigenvectors are found.
   Note that the m matrix is ruined on return from eigen (overwriten by the
   transformation matrix used to transform the matrix into a tridiagonal
   matrix), so you should copy it if you need it after returning from eigen.
   after returning from eigen, m should be freed as it serves no use to
   anyone.
*/

#ifdef TIMES
#include <sys/types.h>
#include <sys/times.h>
#include <stdio.h>
#endif

eigen(m, s, vals, vecs, n)
	matrix m;
	int s;
	matrix vecs;
	vector vals;
	int n;
{
	matrix t;
	int i,j,k;
#ifdef TIMES
	struct tms buffer[4];
#endif

	t=tridiagalloc(s);
#ifdef TIMES
	times(buffer);
#endif
	householder(m,s,t);
#ifdef TIMES
	times(buffer+1);
	fprintf(stderr,"%.2f CPU user seconds used for householder.\n",
		(buffer[1].tms_utime-buffer[0].tms_utime)/60.);

#endif
#ifdef MATLAB
/****/	writetrid((long)s,"trid",t);
/****/	writemat((long)s,(long)s,"ttrans",m);
#endif
	findroots(t,s,n,vals,vecs);
#ifdef TIMES
	times(buffer+2);
	fprintf(stderr,"%.2f CPU user seconds used for findroots.\n",
		(buffer[2].tms_utime-buffer[1].tms_utime)/60.);
#endif

#ifdef MATLAB
/****/  writemat((long)n,(long)s,"tvecs",vecs);
#endif
	/* multiply the found eigenvectors by the transformation matrix in
	   m. use t[0] as a temporary vector.
	*/
	for(i=0;i<n;i++){
		for(j=0;j<s;j++){
			t[0][j]=0;
			for(k=0;k<s;k++)
				t[0][j]+=vecs[i][k]*m[j][k];
		}
		for(j=0;j<s;j++)
			vecs[i][j]=t[0][j];
	}
#ifdef TIMES
	times(buffer+3);
	fprintf(stderr,"%.2f CPU user seconds used for converting eigenvectors.\n",
		(buffer[3].tms_utime-buffer[2].tms_utime)/60.);
#endif
	tridiagfree(t,s);
}
