#include "eigen.h"
#include "stdio.h"

Sturm *sturm_list;

char *malloc();

static sturm_memerr()
{
	fprintf(stderr, "Internal error in sturm.c:\n");
	fprintf(stderr, "\tNot enough memory for object.\n");	
	fprintf(stderr, "Object type to be allocated: ");
	fprintf(stderr,"Sturm data list\n");
	exit(1);
	
}

sturm_free()
{
	Sturm *cursturm,*tmp;

	cursturm=sturm_list;
	while(cursturm){
		tmp=cursturm->next;
		free(cursturm);
		cursturm=tmp;
	}
	sturm_list=0;
}

/* Sturm_gen generates the sturm data for a given tridigonal matrix,
   and puts them in the global list sturm_list, which must be later freed
   with sturm_free.
*/
sturm_gen(m,size)
	matrix m;	/* must be tridiagonal matrix of size size. */
{
	int i;
	Sturm *cursturm;

	if(!(sturm_list=(Sturm*) malloc(sizeof(Sturm))))
		sturm_memerr();
	cursturm=sturm_list;
	cursturm->t.length=1;
	cursturm->bound.lower=0;
	cursturm->flag=0;
	cursturm->t.a=m[0];
	cursturm->t.b=m[1];
	cursturm->next=(Sturm *)0;

	for(i=1;i<size;i++){
		if(ABS(m[1][i-1])<EPSILON){/* if the beta is zero, then
					   it is the start of a new
					   irriducible matrix */
			if(!(cursturm->next=
				(Sturm*) malloc(sizeof(Sturm))))
					sturm_memerr();
			cursturm=cursturm->next;
			cursturm->t.length=1;
			cursturm->bound.lower=0;
			cursturm->flag=0;
			cursturm->t.a=m[0]+i;
			cursturm->t.b=m[1]+i;
			cursturm->next=(Sturm *)0;
			continue;
		}
		cursturm->t.length++;
	}

}

/* If you want to use power smoothing instead of adaptive smoothing, define
   smoothing=POWERSMOOTH
*/
int smoothing=ADAPTIVESMOOTH;

int
sign_changes(sturm, x)
	Sturm sturm;	/* notice that this is a structure, not a pointer */
	scalar x;	/* to it. */
{
	scalar p0,p1,tmp;
	int i,w=0, m;

	scalar smooth;
	if(smoothing==ADAPTIVESMOOTH)
		smooth = (p1=(sturm.t.a[0]-x))!=0 ? ABS(p1) : 1; 
	else
		smooth = ABS(x)>1 ? 1/ABS(x) :  1; 
#ifdef STATS
	s_signchanges++;
	s_schangeslen+=sturm.t.length;
#endif
	p0=1;
	if(smoothing==ADAPTIVESMOOTH){
		if(p1<0){
			p1= -1;
			w++;
			m= -1;
		} else if(p1>0) {
			p1= 1;
			m=1;
		} else {
			p1= 0;
			m=1;
		}
	} else {
		p1=(sturm.t.a[0]-x)*smooth;
		if(p1<0){
			w++;
			m= -1;
		} else
			m=1;
	}

	if(smoothing==ADAPTIVESMOOTH){
		for(i=1;i<sturm.t.length;i++){
			tmp=((sturm.t.a[i]-x)*p1-sturm.t.b[i-1]*sturm.t.b[i-1]*p0/smooth);
			smooth=(tmp!=0) ? ABS(tmp) : 1 ;
			/* We don't do tmp*=smooth now, because we know that tmp will
			   be 1 or -1 or 0. so we only set p1 */
			p0=p1;
			if(tmp<0){
				p1=-1;
				if(m==1)
					w++;
				m= -1;
			} else if (tmp>0){
				p1=1;
				if(m==-1)
					w++;
				m=1;
			} else	p1=0;
		}
	} else {
		for(i=1;i<sturm.t.length;i++){
			tmp=((sturm.t.a[i]-x)*p1-sturm.t.b[i-1]*sturm.t.b[i-1]*p0*smooth)
				*smooth;
			p0=p1;
			if((p1=tmp)<0){
				if(m==1)
					w++;
				m= -1;
			} else if(p1>0){
				if(m==-1)
					w++;
				m=1;
			}
		}
	}
	return w;
}

/* find the bounds on the roots of the characteristic polynomial of the
   irriducible tridiagonal matrix associated with spoly, which is part
   of the matrix given to sturm_gen.
*/
findbounds(psturm)
	Sturm *psturm;
{
	psturm->bound.lower= -1;
	for(;;){
		/* if we are low enough, then stop */
		if(sign_changes(*psturm,psturm->bound.lower)==0)
			break;
		psturm->bound.lower*=2;
	}
	psturm->bound.upper=1;
	for(;;){
		/* if we are high enough, then stop */
		if(sign_changes(*psturm,psturm->bound.upper)==psturm->t.length)
			break;
		psturm->bound.upper*=2;
	}
}


/* find the n'th root of the characteristic polynomial of the irriducible
   tridiagonal matrix associated with spoly, which is part of the matrix
   given to sturm_gen.
*/
scalar
findroot(psturm, n)
	Sturm *psturm;
	int n;
{
	scalar a,b,mid;
	/* the lower bound is always under -1 if already found. if it is
	   0, as initialized in sturm_gen, then we must call findbounds
	   to find bounds
	*/
#ifdef STATS
	s_findroot++;
#endif
	if(!psturm->bound.lower)
		findbounds(psturm);
	
	a=psturm->bound.lower;
	b=psturm->bound.upper;

	/* if STURMEPSILON is given, then we stop at that relative accuracy.
	   if it is not, then find maximum machine accuracy. */
#ifdef STURMEPSILON
	while(((mid=(a+b)/2)-a)>STURMEPSILON*ABS(b) && mid!=b){
#else
	while((mid=(a+b)/2)!=a && mid!=b){
#endif
		if(sign_changes(*psturm,mid)>=(psturm->t.length+1-n)){
			b=mid;
		} else {
			a=mid;
		}
	}
	return mid;
}

/* tsolve solves a tridiagonal, non homogenic systems of equations.
   it solves the system Tv=b, and puts the solution in v.
   note that tsolve destroys the vector b, but does not change the
   trid t.
*/
int
tsolve(t,v,b)
	trid t;
	vector v,b;
{
	int i;
#ifdef STATS
	s_tsolve++;
	s_tsolvelen+=t.length;
#endif
	for(i=0;i<t.length;i++)
		v[i]=t.a[i];
	for(i=t.length-1;i;--i){
		if(v[i]){				/* if not zero */
			v[i-1]-=t.b[i-1]*t.b[i-1]/v[i];
			b[i]/=v[i];
			b[i-1]-=t.b[i-1]*b[i];
		} else {				/* if it is zero */
			printf("GOT TO ZERO CASE IN TSOLVE!\n");
			/* I have to recheck the algorithm in the zero case */
			/* i can't be 1 */
			v[--i]=0;
			/* I am not sure if this is the right thing to do: */
			/* I subtract from every b before the current one
			   the value of x[i-1] (x[n-2] in the explenation I
			   wrote) times the value in the i-1'th column in
			   that row. because this is a tridiagonal matrix,
			   I have to do it only for row i-1 and i-2.
			   the value of x[i-1] is b[i]/b[i-1]. */
			b[i-1]-=t.a[i-1]*(b[i]/b[i-1]);
			b[i-2]-=t.a[i-2]*(b[i]/b[i-1]);
			/* now do the elementary operations, like in the
			   explenation, but also update the b. */
			b[i]/=t.b[i-1];
		}
	}
	/*sum=v[0]=1;*/	/* arbitary */
	if(v[0]==0){
		v[0]=1;
		/*else {
		printf("ERROR: Infinite number of solutions, in tsolve.\n");
		return(0); 
		}*/
	} else
		v[0]=b[0]/v[0];

	for(i=1;i<t.length;i++){
		if(v[i]==0){
			printf("GOT TO ZERO CASE IN TSOLVE!!\n");
			/* i can't be 1 */
			/*i++;
			v[i]= -t.b[i-2]/t.b[i-1]*v[i-2];*/
			v[i]=b[i+1]/t.b[i];
			i++;
			v[i]=b[i]-t.b[i-2]/t.b[i-1]*v[i-2];
		} else {
			/*v[i]= -t.b[i-1]/v[i]*v[i-1];*/
			v[i]=b[i]-t.b[i-1]/v[i]*v[i-1];
		}
	}
	return(1);
}

/* harel takes as an input a irriducible tridiagonal matrix t, a known
   eigenvalue, which was previously found by findroot, and generates the
   respective eigenvector into the output vector, which should have the same
   length as the diagonal of the given irriducible matrix.

   see Numerical Recipes in C, page 394.
*/
/* HAREL_EPSILON is the wanted accuracy of the eigenvectors.
   HAREL_MATITER is the number of iterations we should allow for the reverse
                 iteration algorithm before we choose another initial vector.
   HAREL_MAXRETRYS is the number of times we will try a new initial vector.
*/
#define HAREL_EPSILON	1e-15
#define HAREL_MAXITER	10
#define HAREL_MAXRETRYS	3
/* If the eigenvalue is to accuarate, then we sometimes can't solve the
   resulting equation. So we add to the eigenvalue HAREL_CHANGE times
   that eigenvalue. HAREL_CHANGE was chosen by trying - known problems
   disappeared when choosing this value. If problems do appear then that
   value will have to be increased. Note that if eigenvalues of an irriducible
   matrix are close to one another by a small multiple of HAREL_CHANGE,
   than we may get the incorrect eigenvector...
   Increasing HAREL_CHANGE will also increase the avarage tsolves
   per eigenvector.
*/
#define HAREL_CHANGE	1e-14 /*1e-16*/
harel(t,olambda,v)
	trid t;
	scalar olambda;
	vector v;
{
	vector x[2],na,tmp;
	scalar lambda;
	trid newt;	/* t - lambda*I */ 
	int i, cur=0;	/* x[cur] acts as x0, and x[!cur] acts like y */
	int haderror=0;	/* set to 1 if tsolve has returned an error */
	scalar d,e;

	lambda=olambda+HAREL_CHANGE*olambda;	/* change lambda a little */
	if(ABS(lambda-olambda)<EPSILON){ lambda=EPSILON; }	/* XYZ */

#ifdef STATS
	s_harel++;
	s_harellen+=t.length;
#endif
	x[0]=vectoralloc(t.length);
	x[1]=vectoralloc(t.length);
	tmp=vectoralloc(t.length);
	na=vectoralloc(t.length);

	/* set up newt = t - lambda*I */
	for(i=0;i<t.length;i++)
		na[i]=t.a[i]-lambda;
	newt.length=t.length;
	newt.a=na;
	newt.b=t.b;

newvector:
	if(haderror >= HAREL_MAXRETRYS){
		fprintf(stderr, "ERROR: sturm.c:harel: Giving up finding eigenvector after %d extra trys!\n", HAREL_MAXRETRYS);
		fprintf(stderr, "       Contact programmer.\n", HAREL_CHANGE);
		return;
	}
#ifdef I_DONT_THINK_THIS_WORKS
	if(haderror==0) /*******************/
		for(i=0;i<t.length;i++)
			x[cur][i]=0.0;
	else
#endif
	{
	randomvec(x[cur],t.length);	/* get an initial normalized random
	normalize(x[cur],t.length);	   vector */
	}

	i=HAREL_MAXITER;
	while(i--){
		int j;
		/* copy x[cur] to tmp, because tsolve destroys it */
		for(j=0;j<t.length;j++)
			tmp[j]=x[cur][j];

		if(!(tsolve(newt,x[!cur],x[cur]))){
			haderror++;
			if(haderror==1){
			fprintf(stderr,
		 "Warning: sturm.c:harel: having problems finding eigenvector for %.16g\n",olambda);
			fprintf(stderr,
		 "                        Getting a new random vector\n");
			}
			goto newvector;
		}

		normalize(x[!cur],t.length);

		/* put the difference of the new and previous
		   approximations of the eigenvector in the place of
		   the previous approximation. */
		vvsub(tmp,x[!cur],t.length,x[cur]);
		d=norm(x[cur],t.length);
		vvadd(tmp,x[!cur],t.length,x[cur]);
		e=norm(x[cur],t.length);
		cur=!cur;

		if(d<=HAREL_EPSILON ||  e <= HAREL_EPSILON)
			break;
	}
	if(i==-1){
		haderror++;
		if(haderror==1){
		fprintf(stderr,
	 "Warning: sturm.c:harel: problems finding eigenvector for %.16g\n",olambda);
		fprintf(stderr,
	 "                        Getting a new random vector\n");
		}
		goto newvector;
	}
#if 0
	fprintf(stderr, "Harel stopped after %d iterations.(d=%.16g, e=%16g)\n",HAREL_MAXITER-i,d,e);
#endif
	if(haderror)	/* tell the user that the error is now gone */
		fprintf(stderr, "Notice: sturm.c:harel: Now correctly found eigenvector after %d extra trys.\n", haderror);
	for(i=0;i<t.length;i++)
		v[i]=x[cur][i];

	vectorfree(x[0]);
	vectorfree(x[1]);
	vectorfree(tmp);
	vectorfree(na);
}

/* this is the main routine of this file. as an input it takes a symmetric
   matrix, and it's corresponding size, and it finds the n biggest roots
   of m's characteristic polynomial, and puts it in the vector v, whose
   length must be n. It also finds the corresponding eigen vectors and
   puts them in a matrix which must be n rows and size columns.
*/
findroots(m,size,n,v,vecs)
	matrix m;
	vector v;
	matrix vecs;
{
	int nnoflags=1, nused=0, i, j, k;
	Sturm *s;
	Sturm **vs;	/* for each eigenvalue in v, remember from what sturm
			   it was */
	scalar root;

	sturm_gen(m,size);
	if(!(vs=(Sturm**)malloc(n*sizeof(Sturm *)))) sturm_memerr();

	for(i=1;i<=n && nnoflags!=0; i++){
		nnoflags=0;
		for(s=sturm_list;s;s=s->next){
			if(!s->flag) {
				nnoflags++;
				root=findroot(s,i);
				if(nused<n || root > v[n-1]){
				/* use insertion sort */
					for(j=MIN(n-1,nused);j>=1;j--){
						if(root > v[j-1]){
							v[j]=v[j-1];
							vs[j]=vs[j-1];
						} else {
							v[j]=root;
							vs[j]=s;
							break;
						}
					}
					if(j==0){
						v[0]=root;
						vs[0]=s;
					}
					if(nused<n) nused++;
				} else {
					s->flag=1;
				}
				if(s->t.length==i){
					s->flag=1;
				}
			}
		}
	}
	for(i=0;i<n;i++){
		k=vs[i]->t.a-m[0];
		for(j=0;j<k;j++)
			vecs[i][j]=0;
		harel(vs[i]->t,v[i],vecs[i]+k);
		for(j=k+vs[i]->t.length;j<size;j++)
			vecs[i][j]=0;
	}
		
	free(vs);
	sturm_free();
}
